
# include <stdlib.h>
# include <SR/sr_geo2.h>
# include <SE/se_mesh.h>
# include <SE/se_triangulator.h>

//******************************************************************
//  Used Mesh Types
//******************************************************************

# include "setut_cmesh.h"
typedef CmVertex MyVertex;
typedef CmEdge   MyEdge;
typedef CmFace   MyFace;

//******************************************************************
//  My Triangulator Manager
//******************************************************************

class MyCdtManager: public SeTriangulatorManager
 { public :
    virtual void get_vertex_coordinates ( const SeVertex* v, double& x, double & y );
    virtual void set_vertex_coordinates ( SeVertex* v, double x, double y );
    virtual bool is_constrained ( SeEdge* e );
    virtual void set_unconstrained ( SeEdge* e );
    virtual void get_constraints ( SeEdge* e, SrArray<int>& ids );
    virtual void add_constraints ( SeEdge* e, const SrArray<int>& ids );
 };

void MyCdtManager::get_vertex_coordinates ( const SeVertex* v, double& x, double & y )
 {
   x = (double) ((MyVertex*)v)->c[0];
   y = (double) ((MyVertex*)v)->c[1];
 }

void MyCdtManager::set_vertex_coordinates ( SeVertex* v, double x, double y )
 {
   ((MyVertex*)v)->c[0] = (float) x;
   ((MyVertex*)v)->c[1] = (float) y;
 }

bool MyCdtManager::is_constrained ( SeEdge* e )
 {
   return ((MyEdge*)e)->is_constrained();
 }

void MyCdtManager::set_unconstrained ( SeEdge* e )
 {
   ((MyEdge*)e)->constraints_ids.size(0);
 }

void MyCdtManager::get_constraints ( SeEdge* e, SrArray<int>& ids )
 {
   ids = ((MyEdge*)e)->constraints_ids;
 }

void MyCdtManager::add_constraints ( SeEdge* e, const SrArray<int>& ids )
 {
   for ( int i=0; i<ids.size(); i++ ) ((MyEdge*)e)->constraints_ids.push(ids[i]);
 }

//******************************************************************
//  Ready to insert points in a triangulation
//******************************************************************

// Lets make life easier with global pointers:
MyMesh*         TheMesh;
MySymEdge*      TheBorder;
SeTriangulator* TheTriangulator;

// Global array saving inserted data to be able to rebuild constrained
// and conforming triangulations for comparison purposes
bool ChangingMode=false;
SrArray<float> Vertices;
SrArray<float> Constraints;

static void insert_vertex ( double x, double y )
 { 
   SeBase* s;
   SeTriangulator::LocateResult res;

   printf ( "Vertex %d (%+6.4f,%+6.4f): ", (TheMesh->vertices()-4)+1, x, y );
   res = TheTriangulator->locate_point ( TheBorder->sym()->fac(), x, y, s );

   if ( res==SeTriangulator::NotFound )
    { // within this demo program, this NotFound case should never happen
      printf ( "Error: Not in domain, or faces not ccw!\n" );
    }
   else if ( res==SeTriangulator::VertexFound )
    { printf ( "Duplicated vertex, so not inserted.\n" );
    }
   else if ( res==SeTriangulator::EdgeFound )
    { TheTriangulator->insert_point_in_edge ( s->edg(), x, y );
      printf ( "Inserted in an edge.\n" );
      Vertices.push()=(float)x;
      Vertices.push()=(float)y;
    }
   else // a face was found
    { TheTriangulator->insert_point_in_face ( s->fac(), x, y );
      printf ( "Inserted in a triangle.\n" );
      Vertices.push()=(float)x;
      Vertices.push()=(float)y;
    }
 }

static SeVertex* insert_or_find_vertex ( float x, double y )
 { 
   SeVertex *v=0;
   SeBase *s;
   SeTriangulator::LocateResult res;

   res = TheTriangulator->locate_point ( TheBorder->sym()->fac(), x, y, s );
   if ( res==SeTriangulator::NotFound )
    { // within this demo program, this NotFound case should never happen
      printf ( "Error: Not in domain, or faces not ccw!\n" );
    }
   else if ( res==SeTriangulator::VertexFound )
    { if ( !ChangingMode ) printf ( "Using existing vertex.\n" );
      v = s->vtx();
    }
   else if ( res==SeTriangulator::EdgeFound )
    { v = TheTriangulator->insert_point_in_edge ( s->edg(), x, y );
    }
   else // a face was found
    { v = TheTriangulator->insert_point_in_face ( s->fac(), x, y );
    }
   return v;
 }

static void insert_constraint ( float x1, float y1, float x2, float y2, int id )
 { 
   SeVertex *v1, *v2;

   v1 = insert_or_find_vertex ( x1, y1 );
   if ( !v1 ) return;
   v2 = insert_or_find_vertex ( x2, y2 );
   if ( !v2 ) return;

   TheTriangulator->insert_line_constraint ( v1, v2, id );

   if ( !ChangingMode )
    { Constraints.push()=x1;
      Constraints.push()=y1;
      Constraints.push()=x2;
      Constraints.push()=y2;
      printf ( "Inserted Constraint %d (%+6.4f,%+6.4f):(%+6.4f,%+6.4f).\n", id, x1, y1, x2, y2 );
    }
 }

// Create an initial triangulated square (sides are larger than the window):
static void init ( SeTriangulator* t )
 {
   double xmin, ymin, xmax, ymax;
   xmax=ymax=1.5; xmin=ymin=-1.5;
   TheBorder = (MySymEdge*) TheTriangulator->init_as_triangulated_square
                            ( xmin, ymin, xmax, ymin, xmax, ymax, xmin, ymax )->sym();
 }

static void change_triangulator_mode ()
 {
   int i;
   SeTriangulator::Mode mode = TheTriangulator->mode();
   if ( mode==SeTriangulator::ModeConforming )
    { mode = SeTriangulator::ModeConstrained;
      printf("Switching to Constrained mode.\n");
    }
   else
    { mode = SeTriangulator::ModeConforming;
      printf("Switching to Conforming mode.\n");
    }

   TheTriangulator->mode ( mode );
   init ( TheTriangulator );

   ChangingMode = true;

   for ( i=0; i<Vertices.size(); i+=2 )
    insert_vertex ( Vertices[i], Vertices[i+1] );

   for ( i=0; i<Constraints.size(); i+=4 )
    insert_constraint ( Constraints[i], Constraints[i+1], Constraints[i+2], Constraints[i+3], i/4 );

   ChangingMode = false;
 }

static double random_coord ()
 {
   double n = double(rand()) / double(RAND_MAX); // n E [0,1]
   n = n*1.8f -0.9f;

   return n;
 }

//******************************************************************
//  Draw the mesh with fltk/opengl
//******************************************************************

# include <FL/Fl_Gl_Window.H>
# include <FL/gl.h>
# include <FL/Fl.H>

class TcViewer : public Fl_Gl_Window 
 { public : 
    TcViewer ( int x, int y, int w, int h, const char *label=0 )
         :Fl_Gl_Window ( x, y, w, h, label )
     { Fl::gl_visual ( FL_RGB8 | FL_DOUBLE | FL_DEPTH );
       resizable ( this );
     }

    virtual void draw ();
    virtual int handle ( int event );
 };

void TcViewer::draw ()
 {
   MySymEdge *s;
   MyEdge *e, *ei;

   if ( !valid() )
    { glViewport ( 0, 0, w(), h() );
      glDisable ( GL_DEPTH_TEST ); // to make sure the order in drawing is taken!
      glEnable ( GL_LINE_SMOOTH );
      glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
      glClearColor ( 1, 1, 1, 0 );
    }

   if ( TheMesh->empty() ) return;

   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // draw edges:
   glLineWidth ( 0.5f );
   glBegin ( GL_LINES );

   // we will draw constrained edges after to ensure they are
   // well visible.
   static SrArray<MyEdge*> cedges;

   // draw non-constrained edges:
   glColor3f ( 0, 0, 0.8f );     
   e = ei = TheMesh->first()->edg();
   do { if ( e->is_constrained() )  
         { cedges.push()=e; }
        else 
         { s = e->se();
           glVertex2fv ( s->vtx()->c );
           glVertex2fv ( s->nxt()->vtx()->c );
         }
        e = e->nxt();
      } while ( e!=ei );

   // draw constrained edges:
   glColor3f ( 1, 0, 0 ); 
   while ( cedges.size() )
    { s = cedges.pop()->se();
      glVertex2fv ( s->vtx()->c );
      glVertex2fv ( s->nxt()->vtx()->c );
    }

   glEnd ();
 }

int TcViewer::handle ( int event )
 {
   int x = Fl::event_x();
   int y = Fl::event_y();
   int key = Fl::event_key();

   if ( event==FL_KEYBOARD )
    { switch ( key )
       { case FL_Escape : if ( TheMesh->vertices()==4 ) key=0;
                           else { ::init(TheTriangulator); Vertices.size(0); Constraints.size(0); }
                          break;
         case FL_Enter : insert_vertex ( random_coord(), random_coord() ); break;
         case 32 : change_triangulator_mode(); break; // space
         default : key=0;
       }
      if ( key ) { redraw (); return 1; }
    }
   else if ( event==FL_PUSH )
    { static int count=0;
      static float lpx;
      static float lpy;
      int w = Fl_Window::w();
      int h = Fl_Window::h();
      float px = 2.0f* float(x)/w -1.0f;
      float py = 2.0f* float(y)/h -1.0f;
      py *= -1;

      if ( Fl::event_state(FL_BUTTON3) )
       { insert_vertex ( px, py );
       }
      else if ( Fl::event_state(FL_BUTTON1) )
       { count++;
         if ( count%2==0 ) insert_constraint ( lpx, lpy, px, py, count/2 );
         lpx=px; lpy=py;
       }
      redraw ();
      return 1;
    }
   return 0;
 }

void cdt_main ()
 {
   printf ("Press:\n");
   printf ("  Enter for a random point insertion\n" );
   printf ("  Esc to init triangulation or exit\n" );
   printf ("  Two consecutive left button clicks to add a constraint\n" );
   printf ("  Right mouse button to insert a point\n" );
   printf ("  Space to switch conforming/constrained mode\n" );
   printf ("    (conforming mode is less robust w/intersections)\n" );

   // Init used mesh and triangulator types:
   TheMesh = new MyMesh;
   TheTriangulator = new SeTriangulator
                      ( SeTriangulator::ModeConstrained, // can be converted to conforming on line
                        TheMesh,                         // my mesh to triangulate
                        new MyCdtManager,             // my manager says how to access my mesh
                        0.0001                           // the used epsilon in geometric primitives
                      );

   // Create an initial triangulated square (sides are larger than the window):
   init ( TheTriangulator );

   // Initialize fltk and show the window:
   TcViewer* viewer = new TcViewer ( 200, 200, 400, 300 );
   viewer->label ( "SE CDT Example - M. Kallmann 2003" );
   viewer->show();

   Fl::run();
 }

