
# include <math.h>
# include <stdlib.h>

# include <SR/sr_geo2.h>
# include <SE/se_mesh.h>
# include <SE/se_triangulator.h>

//******************************************************************
//  Used Mesh Types
//******************************************************************

# include "setut_vmesh.h"

typedef VmVertex MyVertex;
typedef VmEdge   MyEdge;
typedef VmFace   MyFace;

//******************************************************************
//  My Triangulator Manager
//******************************************************************

class MyDtManager: public SeTriangulatorManager
 { public :
    virtual void get_vertex_coordinates ( const SeVertex* v, double& x, double & y );
    virtual void set_vertex_coordinates ( SeVertex* v, double x, double y );
 };

void MyDtManager::get_vertex_coordinates ( const SeVertex* v, double& x, double & y )
 {
   x = (double) ((MyVertex*)v)->c[0];
   y = (double) ((MyVertex*)v)->c[1];
 }

void MyDtManager::set_vertex_coordinates ( SeVertex* v, double x, double y )
 {
   ((MyVertex*)v)->c[0] = (float) x;
   ((MyVertex*)v)->c[1] = (float) y;
 }

//******************************************************************
//  Ready to insert points in a triangulation
//******************************************************************

// Lets make life easier with global pointers:
static MyMesh*         TheMesh;
static MySymEdge*      TheBorder;
static SeTriangulator* TheTriangulator;

void insert ( double x, double y )
 { 
   SeBase* s;
   SeTriangulator::LocateResult res;

   printf ( "Vertex %d (%+6.4f,%+6.4f): ", (TheMesh->vertices()-4)+1, x, y );
   res = TheTriangulator->locate_point ( TheBorder->sym()->fac(), x, y, s );

   if ( res==SeTriangulator::NotFound )
    { // within this demo program, this NotFound case should never happen
      printf ( "Error: Not in domain, or faces not ccw!\n" );
    }
   else if ( res==SeTriangulator::VertexFound )
    { printf ( "Duplicated vertex, so not inserted.\n" );
    }
   else if ( res==SeTriangulator::EdgeFound )
    { TheTriangulator->insert_point_in_edge ( s->edg(), x, y );
      printf ( "Inserted in an edge.\n" );
    }
   else // a face was found
    { TheTriangulator->insert_point_in_face ( s->fac(), x, y );
      printf ( "Inserted in a triangle.\n" );
    }
 }

static double random_coord ()
 {
   double n = double(rand()) / double(RAND_MAX); // n E [0,1]
   n = n*1.8f -0.9f;

   return n;
 }

static void insert_spaced_points ( int n, double radius )
 {
   SeBase *sr;
   SeTriangulator::LocateResult res;
   MyVertex* v;
   int nr=0;
   double px, py, vx, vy, dist;
   while ( n-->0 )
    { px = random_coord();
      py = random_coord();
      res = TheTriangulator->locate_point ( TheBorder->sym()->fac(), px, py, sr );

      v = 0;
      if ( res==SeTriangulator::TriangleFound )//|| res==SeTriangulator::EdgeFound )
        v = (MyVertex*) TheTriangulator->insert_point_in_face ( sr->fac(), px, py );
      else if ( res==SeTriangulator::EdgeFound )
        v = (MyVertex*) TheTriangulator->insert_point_in_edge ( sr->edg(), px, py );

      if ( v )
       { vx = v->c[0];
         vy = v->c[1];
         MySymEdge *si, *s, *x;
         si = s = v->se();
         bool pointok=true;
         do { x = s->nxt();
              px=x->vtx()->c[0]; py=x->vtx()->c[1];
              dist = sqrt(sr_dist2(px,py,vx,vy));
              if ( dist<radius ) { pointok=false; break; }
              s = s->rot();
            } while ( s!=si );
         if (!pointok) 
          { TheTriangulator->remove_vertex ( v );
            nr++;
          }
       }
    }
   printf ( "Vertices in the mesh=%d  Rejected=%d\n", TheMesh->vertices(), nr );
 }

//******************************************************************
//  Draw the mesh with fltk/opengl
//******************************************************************

# include <FL/Fl_Gl_Window.H>
# include <FL/gl.h>
# include <FL/Fl.H>

class TpViewer : public Fl_Gl_Window 
 { public : 
    TpViewer ( int x, int y, int w, int h, const char *label=0 )
         :Fl_Gl_Window ( x, y, w, h, label )
     { Fl::gl_visual ( FL_RGB8 | FL_DOUBLE | FL_DEPTH );
       resizable ( this );
     }

    virtual void draw ();
    virtual int handle ( int event );
 };

void TpViewer::draw ()
 {
   MySymEdge *s;
   MyEdge *e, *ei;

   if ( !valid() )
    { glViewport ( 0, 0, w(), h() );
      glDisable ( GL_DEPTH_TEST ); // to make sure the order in drawing is taken!
      glEnable ( GL_LINE_SMOOTH );
      glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
      glClearColor ( 1, 1, 1, 0 );
    }

   if ( TheMesh->empty() ) return;

   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // draw edges:
   glLineWidth ( 0.5f );
   glBegin ( GL_LINES );
   glColor3f ( 0, 0, 0.8f );
     
   // loop over the circular list of edges:
   e = ei = TheMesh->first()->edg();
   do { s = e->se();
        glVertex2fv ( s->vtx()->c );
        glVertex2fv ( s->nxt()->vtx()->c );
        e = e->nxt();
      } while ( e!=ei );

   glEnd ();
 }

int TpViewer::handle ( int event )
 {
   int x = Fl::event_x();
   int y = Fl::event_y();
   int key = Fl::event_key();

   if ( event==FL_KEYBOARD )
    { switch ( key )
       { case FL_Enter : ::insert ( random_coord(), random_coord() ); break;
         case 32 : ::insert_spaced_points ( 10000, 0.05 ); break; // space
         default : key=0;
       }
      if ( key ) { redraw (); return 1; }
    }
   else if ( event==FL_PUSH )
    { double w = Fl_Window::w();
      double h = Fl_Window::h();
      double fx = 2.0f* double(x)/w -1.0f;
      double fy = 2.0f* double(y)/h -1.0f;
      ::insert ( fx, fy*-1.0f );
      redraw ();
      return 1;
    }
   return 0;
 }

void dt_main ()
 {
   printf ("Press Enter for a random insertion\n" );
   printf ("Mouse click for point insertion\n" );
   printf ("Space for inserting several points w/ minimal distance\n" );

   // Init used mesh and triangulator types:
   TheMesh = new MyMesh;
   TheTriangulator = new SeTriangulator
                      ( SeTriangulator::ModeUnconstrained, // no constraints are used here
                        TheMesh,                           // my mesh to triangulate
                        new MyDtManager,                   // my manager says how to access my mesh
                        0.0001                             // the used epsilon in geometric primitives
                      );

   // Create an initial triangulated square (sides are larger than the window):
   double xmin, ymin, xmax, ymax;
   xmax=ymax=1.5; xmin=ymin=-1.5;
   TheBorder = (MySymEdge*) TheTriangulator->init_as_triangulated_square
                            ( xmin, ymin, xmax, ymin, xmax, ymax, xmin, ymax )->sym();

   // Initialize fltk and show the window:
   TpViewer* viewer = new TpViewer ( 200, 200, 400, 300 );
   viewer->label ( "SE DT Example - M. Kallmann 2003" );
   viewer->show();

   Fl::run();
 }

