
# include <FL/Fl.H>
# include <FL/fl_draw.H>

# include <math.h>

# include <SR/sr.h>
# include <SR/sr_swing_control.h>
# include <SR/sr_alg.h>
# include <SR/sr_output.h>

//============================ SrSwingControl ============================

# define SWMIN -srpi
# define SWMAX srpi

SrSwingControl::SrSwingControl ( int x, int y, int w, int h, const char *l )
               :Fl_Widget ( x, y, w, h, l )
 {
   _sx = _sy = 0;
   _ex = _ey = srpi;
   _swap = true;
   _crossr = 3;
 }

static float flinear ( double val, double smin, double smax, double gmin, double gmax )
{
  if (smin == smax) return (float)gmax;
  else return float( gmin + (gmax - gmin) * (val - smin) / (smax - smin) );
}

bool SrSwingControl::value ( float x, float y )
 {
   clear_changed();
   if ( x==_sx && y==_sy ) return false;

   // test length of the axis-angle
   float n = sqrtf(x*x+y*y);
   if ( n>srpi ) { n=srpi/n; x*=n; y*=n; } // make length to be PI
   
   // test ellipse limits
   if ( sr_in_ellipse(x,y,_ex,_ey)>0 )
    { sr_get_closest_on_ellipse ( _ex, _ey, x, y );
    }

   // set value:
   _sx = x;
   _sy = y;
   set_changed();
   redraw();
    
   return true;
 }

void SrSwingControl::ellipse ( float ex, float ey )
 {
   if ( _ex==ex && _ey==ey ) return;
   _ex = SR_BOUND(ex,0.0001f,srpi);
   _ey = SR_BOUND(ey,0.0001f,srpi);
   value ( _sx, _sy );
   set_changed();
   redraw();
 }

void SrSwingControl::draw()
 {
   # define GETX(xx) int(flinear(xx,SWMIN,SWMAX,x1,xe)+0.5f)
   # define GETY(yy) int(flinear(yy,SWMIN,SWMAX,y1,ye)+0.5f)
   # define SETRED   fl_color(255,0,0)
   # define SETGREEN fl_color(0,255,0)

   int x1 = x() + 4;
   int y1 = y() + 4;
   int w1 = w() - 2 * 4;
   int h1 = h() - 2 * 4;
   int xe = x1+w1-1;
   int ye = y1+h1-1;
   int xi, yi;

   draw_box ( box(), x(), y(), w(), h(), color() );

   // draw axis
   xi = GETX(0);
   yi = GETY(0);
   
   if (_swap) SETRED; else SETGREEN;
   fl_yxline ( xi, y1, y1+h1 );
   if (_swap) SETGREEN; else SETRED;
   fl_xyline ( x1, yi, x1+w1 );

   // draw ellipse
   float x, y, a;
   fl_color ( 90, 90, 90 );
   fl_begin_loop();
   for ( a=0; a<=sr2pi; a+=0.1f )
    { x = _ex * cosf(a);
      y = _ey * sinf(a);
      if ( _swap ) { float tmp; SR_SWAP(x,y); }
      fl_vertex ( GETX(x), GETY(y) );
    }
   fl_end_loop();
   
   // draw segments
   int i;
   SrPnt2 p1, p2;
   fl_color(0,0,255);
   for ( i=0; i<_points.size()-1; i+=2 )
    { p1 = _points[i];
      p2 = _points[i+1];
      if ( _swap ) { float tmp; SR_SWAP(p1.x,p1.y); SR_SWAP(p2.x,p2.y); }
      fl_line ( GETX(p1.x), GETY(p1.y), GETX(p2.x), GETY(p2.y) );
    }

   // draw value
   if ( _swap )
    { xi = GETX(_sy); yi = GETY(_sx); }
   else
    { xi = GETX(_sx); yi = GETY(_sy); }
   fl_color ( 90, 90, 90 );
   fl_yxline ( xi, yi-_crossr, yi+_crossr );
   fl_xyline ( xi-_crossr, yi, xi+_crossr );

   # undef GETX
   # undef GETY
   # undef SETRED
   # undef SETGREEN
 }

int SrSwingControl::handle ( int e )
 {
   int ret = 0;
   
   switch ( e )
    { case FL_PUSH:
      case FL_DRAG:
       { float x1 = (float)x() + 4;
         float y1 = (float)y() + 4;
         float w1 = (float)w() - 2 * 4;
         float h1 = (float)h() - 2 * 4;
         float x = flinear(Fl::event_x(), x1, x1+w1-1.0, SWMIN, SWMAX);
         float y = flinear(Fl::event_y(), y1, y1+h1-1.0, SWMIN, SWMAX);
         if ( _swap ) { float tmp; SR_SWAP(x,y); }
         if ( value(x,y) ) do_callback();
         ret = 1;
         Fl::focus(this);
         set_visible_focus(); // not sure what this does
       } break;
       
      case FL_KEYBOARD:
       { float incx=0, incy=0, inc=0.005f;
         if ( Fl::event_state(FL_SHIFT) ) inc*=4;
         if ( Fl::event_state(FL_CTRL) ) inc/=10;
         switch ( Fl::event_key() )
          { case FL_Left  : incx-=inc; ret=1; break;
            case FL_Right : incx+=inc; ret=1; break;
            case FL_Up    : incy+=inc; ret=1; break;
            case FL_Down  : incy-=inc; ret=1; break;
          }
         if ( ret )
          { if ( _swap ) { float tmp; SR_SWAP(incx,incy); incx*=-1; incy*=-1; }
            if ( value ( _sx+incx, _sy-incy ) ) do_callback();
          }
       } break;
    }
    
   return ret;
}

//================================ End of File =================================================
