
# include <FL/fl_ask.h>

//# include <SR/sr_event.h>
//# include <SR/sr_plane.h>
//# include <SR/sr_camera.h>
# include <SR/sr_lines.h>
# include <SR/sr_model.h>
# include <SR/sr_sn_group.h>
# include <SR/sr_sn_shape.h>
# include <SR/sr_sn_manipulator.h>

# include "srm_material_tab.h"
# include "srm_main_win.h"
# include "srm_main.h"
# include "srm_viewer.h"

# define SR_USE_TRACE1  // event tracing
# include <SR/sr_trace.h>

SrmViewer::SrmViewer ( int x, int y, int w, int h ) : SrViewer(x,y,w,h) 
 {
   _lastfac = _lastmtl = -1;
   _root = new SrSnGroup;

   SrViewer::root ( _root );
 }

void SrmViewer::add_model ( SrSnModel* model )
 {
   SrSnManipulator* manip = new SrSnManipulator;
   SrSnGroup* group = new SrSnGroup;
   SrSnLines* lines = new SrSnLines;

   lines->visible ( false );     
   manip->visible ( false );

   group->add ( model );
   group->add ( lines );
   manip->child ( group );

   _root->add ( manip );
   
   SrViewer::view_all();
 }

bool SrmViewer::remove_model ( int i )
 {
   if ( i<0 || i>=_root->size() ) return false;
   _root->remove ( i );
   return true;
 }

SrSnModel* SrmViewer::get_smodel ( int i )
 {
   if ( i<0 || i>=_root->size() ) return 0;

   SrSnManipulator* manip = (SrSnManipulator*)_root->get(i);
   SrSnGroup* group = (SrSnGroup*)manip->child();
   return (SrSnModel*)group->get(0);
 }

SrModel* SrmViewer::get_model ( int i )
 {
   SrSnModel* smodel = get_smodel ( i );
   if ( !smodel ) return 0;
   return &smodel->shape();
 }

SrSnManipulator* SrmViewer::get_manipulator ( int i )
 {
   if ( i<0 || i>=_root->size() ) return 0;
   SrSnManipulator* manip = (SrSnManipulator*)_root->get(i);
   return manip;
 }

void SrmViewer::reset_normals ( int i )
 {
   if ( i<0 || i>=_root->size() ) return;

   SrSnManipulator* manip = (SrSnManipulator*)_root->get(i);
   SrSnGroup* group = (SrSnGroup*)manip->child();
   SrSnLines* slines = (SrSnLines*) group->get(1);
   SrLines& l = slines->shape();

   l.init ();
   if ( slines->visible() ) make_normals (i);
 }

# define ADDNORM(c,n) l.push_line ( c, c+n*nsize )

void SrmViewer::make_normals ( int i )
 {
   if ( i<0 || i>=_root->size() ) return;

   SrSnManipulator* manip = (SrSnManipulator*)_root->get(i);
   SrSnGroup* group = (SrSnGroup*)manip->child();
   const SrModel& m = ((SrSnModel*)group->get(0))->const_shape();
   SrLines& l = ((SrSnLines*)group->get(1))->shape();

   l.init ();

   SrBox box;
   m.get_bounding_box(box);
   float nsize = box.max_size()/20.0f;

   int f, a, b, c, an, bn, cn;
   SrPnt center;
   SrVec normal;

   l.push_color ( SrColor::red );

   if ( m.N.empty() ) // there are no normals in N
    { for ( f=0; f<m.F.size(); f++ )
       { a = m.F[f].a; b=m.F[f].b; c=m.F[f].c;
         center = ( m.V[a] + m.V[b] + m.V[c] )/3.0f;
         ADDNORM ( center, m.face_normal(f) );
       }
    }
   else if ( m.Fn.size()>0 ) // using normals in Fn
    { for ( f=0; f<m.F.size() && f<m.Fn.size(); f++ )
       { a = m.F[f].a; b=m.F[f].b; c=m.F[f].c;
         an = m.Fn[f].a; bn=m.Fn[f].b; cn=m.Fn[f].c;
	     ADDNORM ( m.V[a], m.N[an] );
 	     ADDNORM ( m.V[b], m.N[bn] );
 	     ADDNORM ( m.V[c], m.N[cn] );
       }
    }
 }

# undef ADDNORM

void SrmViewer::set_visibility ( int i, bool model, bool normals, bool manip )
 {
   if ( i<0 || i>=_root->size() ) return;

   //sr_out<<model<<srspc<<normals<<srspc<<manip<<srnl;
   SrSnManipulator* smanip = (SrSnManipulator*)_root->get(i);
   SrSnGroup* group = (SrSnGroup*) smanip->child();
   SrSnModel* smodel = (SrSnModel*) group->get(0);
   SrSnLines* slines = (SrSnLines*) group->get(1);

   smodel->visible(model);
   slines->visible(normals);
   smanip->visible(manip);

   if ( normals && slines->const_shape().empty() ) make_normals ( i );
//   if ( manip ) smanip->update();
  }

void SrmViewer::get_visibility ( int i, bool& model, bool& normals, bool& manip )
 {
   model = normals = manip = false;

   if ( i<0 || i>=_root->size() ) return;

   SrSnManipulator* smanip = (SrSnManipulator*)_root->get(i);
   SrSnGroup* group = (SrSnGroup*) smanip->child();
   SrSnModel* smodel = (SrSnModel*) group->get(0);
   SrSnLines* slines = (SrSnLines*) group->get(1);

   model = smodel->visible();
   normals = slines->visible();
   manip = smanip->visible();
   //sr_out<<smodel->shape().name<<": "<<model<<srspc<<normals<<srspc<<manip<<srnl;
 }

//==================== virtuals ===================================

int SrmViewer::handle_keyboard ( const SrEvent& e )
 {
   return SrViewer::handle_keyboard(e);
 }

int SrmViewer::handle_scene_event ( const SrEvent& e )
 {
   if ( e.type==SrEvent::Push &&
        TheMainWin->ui_pick_face_but->value() &&
        TheMaterialTab->cur_color()>=0 &&
        TheMainWin->cur_model() )
    {
      SrModel& m = *TheMainWin->cur_model();
      int f = m.pick_face ( e.ray );
      if ( f>=0 )
       { _lastmtl=m.Fm[f]; _lastfac=f; // save here in case of undo
         m.Fm[f] = TheMaterialTab->cur_color();
         TheMainWin->cur_smodel()->changed(true);
         TheMainWin->ui_viewer->render();
       }
      return 1;
    }
   return SrViewer::handle_scene_event(e);
 }

//============================= EOF ===================================
