
/* The class SrArrayObj is a class to maintain an array of objects respecting
   the elements constructors, destructors, copy operators, etc. For this it
   is fully written as a template, and cannot have the derivation solution used
   by SrArrayBase and SrArray.
   SrArrayObj is not included in the Sr toolkit because SrArray and SeArrayPt
   can give the same functionality but with better performance due to the
   direct memory management rountines in SrArray. 
   Anyway, as SrArrayObj might be easier to use in some cases, it is included
   here as a test example, together with SrQueueObj.
*/

# ifndef SR_ARRAY_OBJ_H
# define SR_ARRAY_OBJ_H

/** \file sr_array_obj.h 
 * resizeable array template for classes */

# include <SR/sr.h> 
# include <SR/sr_output.h> 

/*! \class SrArrayObj sr_array_obj.h
    \brief Resizeable array of classes

    SrArrayObj can be used only with classes that provides the needed copy constructors,
    constructors, destructors, etc. If these methods are correctly implemented, all resize
    functions of SrArrayObj will correctly work. In particular, class X has to implement
    the method void take_data ( X& x ), in order to optimize possible internal memory
    managements of class X. This method should make the class to contain the same data of
    the given x parameter, but x can be modified. To adjust the array size when needed, 
    a gap value is kept, see the description of each method for an explanation of how
    it is used. 
    Before, a commom SrArrayBaseTpl class was used containing the commom
    methods of SrArray and SrArrayObj, but some hacks with void* pointers were necessary 
    to let the two work properly, and then this approach was abandoned. */
template <class X>
class SrArrayObj
 { private :
    X*  _data;        //!< Array pointer used for storage
    int _size;        //!< Number of elements being used in array
    int _capacity;    //!< Number of allocated elements (>=size)
    int _gap;         //!< Gap to use in methods like insert(), and insort()
   public :

    /*! Constructs allocating space for the given size, that has a default value of 0. 
        The given size must have a >=0 value, and gap is set to 8. */
    SrArrayObj ( int s=0 ) { _data=s>0?new X[s]:0; _size=_capacity=s; _gap=8; }

    /*! Constructs with the given size, capacity and gap. Gap has a default value of 8.
      The given size must have a >=0 value the given capacity must be >= the given size. */
    SrArrayObj ( int s, int c, int g=8 ) { _data=c>0?new X[c]:0; _size=s; _capacity=c; _gap=g; }

    /*! Copy constructor. SrArrayObj will be an exact copy of the given array. 
        Allocating the same capacity of a, but copying only a.size() elements. 
        The operator=() of SrArrayObj is used. */
    SrArrayObj ( const SrArrayObj& a ) { _data=0; (*this)=a; }

    /*! Constructor from a given buffer. No checkings are done, its the user 
        responsability to give consistent parameters. */
    SrArrayObj ( X* pt, int s, int c, int g ) { _data=pt; _size=s; _capacity=c; _gap=g; }

    /*! Destructor deletes the internal array. */
    ~SrArrayObj () { delete [] ((X*)_data); }

    /*! Returns true if the array has no elements, ie, size()==0; and false otherwise. */
    bool empty () const { return _size==0? true:false; }

    /*! Returns the capacity of the array. Capacity is used to be able to have a 
        bigger storage buffer than the current size used. The method capacity() 
        will always return a value not smaller than size(). */
    int capacity () const { return _capacity; }

    /*! Returns the gap being used. Whenever a new element is inserted in the array,
        the new capacity of the internal buffer will be always a multiple of gap. */
    int gap () const { return _gap; }

    /*! Changes the gap being used. Whenever a new element is inserted in the array,
        the new capacity of the internal buffer will be always a multiple of gap. 
        Gap cannot be less than 1, and this is ensured in this method. The constructor
        of SrArray use a default value of 8 to initialize gap. */
    void gap ( int g ) { _gap = g<1? 1:g; }

    /*! Returns the current size of the array. */
    int size () const { return _size; }

    /*! Allows to change the size of the array. In this method, the current value of 
        gap() is not used if reallocation is needed to reach the desired size. If the 
        requested size is 0, or smaller than the capacity, no reallocation is done. */
    void size ( int ns ) { size ( ns, _capacity ); }

    /*! Allows to change the size and the capacity of the array. In this method, the 
        current value of gap() is not used if reallocation is needed to reach the 
        desired capacity. If the size requested is greater than the capacity requested, 
        the capacity will be the same of size. If the requested size is 0, or smaller 
        than the capacity, no reallocation is done. */
    void size ( int ns, int nc );

    /*! Sets all elements as x, using the operator=() */
    void setall ( const X& x ) { for ( int i=0; i<_size; i++ ) _data[i]=x; }

    /*! Makes capacity to be equal to size, deleting all extra capacity if any. */
    void compress ();

    /*! Gets a const reference to the element of index i. Indices start from 0 and must 
        be smaller than size(). No checkings are done to ensure that i is valid. */
    const X& get ( int i ) const { return _data[i]; }

    /*! Operator version of X& get(int i), but returning a non const reference.
        No checkings are done to ensure that i is valid. */
    X& operator[] ( int i ) { return _data[i]; }

    /*! Returns a const pointer of the internal buffer. The internal buffer 
        will always contain a contigous storage space of capacity() elements. 
        See also take_data() and leave_data() methods. */
    operator const X* () const { return _data; }

    /*! Returns a reference to the last element, ie, with index size()-1.
        The array must not be empty when calling this method. */
    X& top () { return _data[_size-1]; }

    /*! Returns a reference to the last element, ie, with index size()-1, and
        then reduces the size of the array by one with no reallocation.
        The array must not be empty when calling this method. */
    X& pop () { return _data[--_size]; }

    /*! Method to append positions. If reallocation is needed, capacity is augmented, 
        keeping it a multiple of gap. The first new element appended is returned as a reference. 
        The quantity of appended positions (dp) has a default value of 1. */
    X& push ( int dp=1 ) { return insert ( _size, dp ); }

    /*! Pushes one position the end of the array using the other push() method, and
      then copies the content of the given element using operator=(). A reference
      to the inserted element is returned. */
    X& push ( const X& x ) { return push()=x; }

    /*! Inserts dp positions, starting at pos i, moving all data correctly. 
        Parameter i can be between 0 and size(), if i==size(), dp positions are appended. 
        If reallocation is needed, capacity is augmented, keeping it a multiple of gap. 
        The first new element inserted (i) is returned as a reference. 
        The quantity of appended positions (dp) has a default value of 1. */
    X& insert ( int i, int dp=1 );

    /*! Removes dp positions starting from pos i, moving all data correctly;
      dp has a default value of 1. */
    void remove ( int i, int dp=1 );

    /*! Copies all internal data of a to SrArrayObj, using the operator=().
        This method has no effect if a "self copy" is called. */
    const SrArrayObj& operator = ( const SrArrayObj<X>& a );

    /*! Inserts the element x in the sorted array, moving all data correctly, and 
        returning the position of the element inserted. When allowdup is false, 
        the element will not be inserted in case of duplication, and in this case, 
        -1 is returned. The compare function int sr_compare(const X*,const X*) must exist.
        Method insert() is called to open the required space, but the operator=() of X
        will be used to copy its contents in the open position.
        Parameter allowdup has a default value of true. */
    int insort ( const X& x, bool allowdup=true );

    /*! Standard library qsort() wrapper call.
        Compare function int sr_compare(const X*,const X*) must exist! */
    void sort ( SR_COMPARE_FUNC )
        { ::qsort ( (void*)_data, (size_t)_size, (size_t)sizeof(X), (srcompare)sr_compare_func ); }

    /*! Linear search, returns index of the element found, or -1 if not found.
        Compare function int sr_compare(const X*,const X*) must exist! */
    int lsearch ( const X& x );

    /*! Binary search for sorted arrays. Returns index of the element found, 
        or -1 if not found. If not found and pos is not 0, pos will have the 
        position to insert the element keeping the array sorted. I tested and 
        this is faster than the standard C library bsearch() for large arrays. 
        Compare function int sr_compare(const X*,const X*) must exist! */
    int bsearch ( const X& x, int *pos=NULL ); 

    /*! Makes the given xpt pointer to point to the internal buffer array, without
        reallocation; xpt may then be null or contain the address of the memory
        used and that was allocated with operator new. After this call, SrArrayObj
        will become an empty valid array (gap is not modified). */
    void leave_data ( X*& xpt ) { xpt=_data; _data=0; _capacity=_size=0; }

    /*! Frees the data of SrArrayObj, and then makes SrArrayObj be the given array a.
        After this, a is set to be a valid empty array. The data is moved without
        reallocation, and a.gap() does not change. */
    void take_data ( SrArrayObj<X>& a );

    /*! Outputs all elements of the array. Element type X must have its ouput operator <<
      available. */
    // friend SrOutput& operator<< ( SrOutput& o, const SrArrayObj<X>& a );
 };

template <class X>
void SrArrayObj<X>::size ( int ns, int nc )
 {
   if ( ns<0  ) ns = 0;
   if ( nc<ns ) nc = ns;
   if ( nc<=0 ) { delete [] _data; _data=0; _capacity=_size=0; return; }

   if ( nc!=_capacity ) 
    { X *newdata = new X[nc];
      for ( int i=0; i<_size && i<ns; i++ ) newdata[i].take_data ( _data[i] );
       delete [] _data; 
      _data = newdata;
      _capacity = nc;
    }
   _size = ns;
 }

template <class X>
void SrArrayObj<X>::compress ()
 {
   if ( _size==_capacity ) return;
   X *newdata = _size? new X[_size] : 0;
   for ( int i=0; i<_size; i++ ) newdata[i].take_data ( _data[i] );
   delete [] _data; 
   _data = newdata;
   _capacity = _size;
 }

template <class X>
X& SrArrayObj<X>::insert ( int i, int dp )
 { 
   int p, ns = _size+dp;

   if ( ns>_capacity ) 
    { _capacity = SR_SIZE_WITH_GAP(ns,_gap);
      X* newdata = new X[_capacity];
      for ( p=0;  p<i; p++ ) newdata[p].take_data ( _data[p] );
      for ( ; p<_size; p++ ) newdata[p+dp].take_data ( _data[p] );
      delete [] _data;
      _data = newdata;
    }
   else 
    { for ( p=ns-1; p>=i+dp; p-- ) _data[p].take_data ( _data[p-dp] );
    }

   _size = ns;
   return _data[i];
 }

template <class X>
void SrArrayObj<X>::remove ( int i, int dp )
 { 
   _size-=dp;
   for ( int p=i; p<_size; p++ ) _data[p].take_data ( _data[p+dp] );
 }

template <class X>
const SrArrayObj<X>& SrArrayObj<X>::operator = ( const SrArrayObj<X>& a )
 { 
   if ( this==&a ) return *this;
   if ( _data ) { delete [] _data; _data=0; }
   _capacity = a._capacity;
   _size = a._size;
   _gap = a._gap;

   if ( _capacity>0  ) 
    { _data = new X [_capacity];
      for ( int i=0; i<_size; i++ ) _data[i]=a[i];
    }
   return *this;
 }

template <class X>
int SrArrayObj<X>::insort ( const X& x, bool allowdup )
 { 
   int result, pos;
   result = bsearch ( x, &pos );
   if ( result<_size && !allowdup ) return _size;
   insert ( pos, 1 ) = x;
   return pos;
 }

template <class X>
int SrArrayObj<X>::lsearch ( const X& x )
 { 
   X *pt = _data;

   for ( int i=0; i<size(); i++ ) 
    { if ( ((int(*)(const X*,const X*))sr_compare)(pt,&x)==0 ) return i; 
      pt++;
    }

   return -1;
 }

template <class X>
int SrArrayObj<X>::bsearch ( const X& x, int *pos )
 {
   int comp;
   int i, e, p;

   comp=1; i=0; e=size(); p=(e-i)/2;
   
   while ( comp!=0 && i!=e )
    { comp = ((int(*)(const X*,const X*))sr_compare) ( &x, _data[p] );
      if ( comp<0 ) e = p;
       else if ( comp>0 ) i = p+1;
      p = i + (e-i)/2;
    }

   if (pos) *pos=p;

   return comp==0? p : -1;
 }

template <class X>
void SrArrayObj<X>::take_data ( SrArrayObj& a )
 {
   if ( _data ) delete [] _data;
   _data     = a._data;     a._data=0;
   _capacity = a._capacity; a._capacity=0;
   _size     = a._size;     a._size=0;
   _gap      = a._gap;
 }

template <class X>
SrOutput& operator<< ( SrOutput& o, const SrArrayObj<X>& a )
 {
   o << '[';
   for ( int i=0; i<a.size(); i++ )
     { o << a[i];
       if ( i<a.size()-1 ) o<<' ';
    }
   return o << ']';
 }

#endif // SR_ARRAY_H


# ifndef SR_QUEUE_OBJ_H
# define SR_QUEUE_OBJ_H

/** \file sr_queue_obj.h 
 * Template for a queue based on SrArrayObj. */

//# include <SR/sr_array_obj.h> 

/*! \class SrQueueObj sr_queue_obj.h
    \brief Queue based on two SrArrayObj

    SrQueueObj implements a FIFO queue using two internal SrArrayObj objects,
    one for elements being inserted, and the other for elements to output.
    Both SrArrayObj work as stacks. See also SrQueue for a speed-optimized
    queue implementation. */
template <class X>
class SrQueueObj
 { private :
    SrArrayObj<X> _in;
    SrArrayObj<X> _out;
   public :

    /*! Constructor receiving the capacity (default is 0), and gap (default 
        is 8) for both input and output arrays. */
    SrQueueObj ( int c=0, int g=8 ) : _in(0,c,g), _out(0,c,g) {}

    /*! Copy constructor. */
    SrQueueObj ( const SrQueueObj& q ) : _in(q._in), _out(q._out) {}

    /*! Returns true if the queue is empty, false otherwise. */
    bool empty () const { return _in.empty() && _out.empty()? true:false; }

    /*! Returns the number of elements in the queue. */
    int size () const { return _in.size()+_out.size(); }

    /*! Compress both input and output arrays. */
    void compress () { _in.compress(); _out.compress(); }

    /*! Returns a reference to the top (first) element; the queue must not be empty. */
    X& top () { return _out.empty()? _in[0]:_out.top(); }

    /*! Returns a reference to the tail (last) element; the queue must not be empty. */
    X& tail () { return _in.empty()? _out[0]:_in.top(); }

    /*! Inserts an element at the end of the queue. */
    X& insert () { return _in.push(); }

    /*! Removes and returns a reference to the first element (top) of the queue. */
    X& remove ()
     { if ( _out.empty() )
        { _out.size(_in.size()-1);
          for ( int i=0; i<_out.size(); i++ ) _out[i]=_in.pop();
          return _in.pop();
        }
       else return _out.pop();
     }

    /*! Empties the in stack, correctly moving its elements to the out stack. */
    void flush ()
     { if ( _in.empty() ) return;
       _out.insert ( 0, _in.size() );
       for ( int i=0; _in.size()>0; i++ ) _out[i]=_in.pop();
     }

    /*! Makes this queue be the same as q, which becomes an empty queue. */
    void take_data ( SrQueueObj<X>& q )
     { _in.take_data ( q._in );
       _out.take_data ( q._out );
     }

    /*! Flushes and then print the elements of the queue that will all be in the output array. */
    friend SrOutput& operator<< ( SrOutput& o, SrQueueObj<X>& q )
     { q.flush(); return o<<q._out; }
 };

#endif // SR_QUEUE_OBJ_H


// ================== SMALL TEST PROGRAM ===========================


# include <stdlib.h>
# include <SR/sr.h>
# include <SR/sr_string.h>

int arrayobj_main ( int, char ** )
 {
   SrArrayObj<SrString> a;
   a.push() = "string10";
   a.push() = "string40";
   a.push() = "string20";
   a.push() = "string05";
   sr_out << a << srnl;
   a.sort ( sr_compare );
   sr_out << a << srnl;

   SrQueueObj<SrString> q;
   q.insert() = "stringA";
   q.insert() = "stringB";
   q.insert() = "stringC";
   q.insert() = "stringD";
   sr_out << q << srnl;

   return 0;
 }

//============================== end of file ===============================
