
# include <stdlib.h>

# include <FL/Fl.H>
# include <FL/fl_file_chooser.H>
# include <FL/fl_ask.H>

# include <SR/sr_timer.h>
# include <SR/sr_model.h>
# include <SR/sr_lines.h>
# include <SR/sr_sn_shape.h>
# include <SR/sr_sn_matrix.h>
# include <SR/sr_sn_manipulator.h>
# include <SR/sr_viewer.h>
# include <SR/sr_bv_tree_query.h>

SrSnManipulator* load ( const char* file )
 {
   SrSnModel* model = new SrSnModel;
   FILE *stream;
   fopen_s(&stream,file,"rt");
   SrInput in ( stream );
   if ( !in.valid() || !model->shape().load(in) )
     sr_out.fatal_error("Could not read model!\n");
   model->shape().set_one_material ( SrMaterial() );
   model->shape().M.push() = SrMaterial();
   model->shape().M.top().diffuse = SrColor::red;
   SrSnManipulator* manip = new SrSnManipulator;
   manip->child ( model );
   return manip;
 }

struct Data
 { SrSnManipulator* manip1;
   SrSnManipulator* manip2;
   SrSnLines* lines;
   SrSnModel* model1;
   SrSnModel* model2;
   SrBvTree* tree1;
   SrBvTree* tree2;
   SrBvTreeQuery query;
   SrTimer timer;
   Data ( SrSnManipulator* m1, SrSnManipulator* m2, SrSnLines* l );
   void measure ();
 };

Data::Data ( SrSnManipulator* m1, SrSnManipulator* m2, SrSnLines* l )
 {
   manip1 = m1;
   manip2 = m2;
   lines = l;
   tree1 = new SrBvTree;
   tree2 = new SrBvTree;
   model1 = (SrSnModel*)manip1->child();
   model2 = (SrSnModel*)manip2->child();
   tree1->make ( model1->shape() );
   tree2->make ( model2->shape() );
 }

void Data::measure ()
 {
   int i, c;

   // get all pairs in collision:
   timer.start();
   c = query.collide_all ( manip1->mat(), tree1, manip2->mat(), tree2 );
   timer.stop();
   float dt1 = (float)timer.dt();

   // true/false collide query:
   timer.start();
   bool b = query.collide ( manip1->mat(), tree1, manip2->mat(), tree2 );
   timer.stop();
   float dt2 = (float)timer.dt();

   // distance query:
   timer.start();
   float dist = query.distance ( manip1->mat(), tree1, manip2->mat(), tree2, 0, 0 );
   timer.stop();
   float dt3 = (float)timer.dt();
   SrPnt dp1 = query.p1()*manip1->mat();
   SrPnt dp2 = query.p2()*manip2->mat();

   // the following tolerance query must always return true:
   timer.start();
   bool tole = query.tolerance ( manip1->mat(), tree1, manip2->mat(), tree2, dist );
   timer.stop();
   float dt4 = (float)timer.dt();
   SrPnt tp1 = query.p1()*manip1->mat();
   SrPnt tp2 = query.p2()*manip2->mat();

   // the following tolerance query must always return true:
   timer.start();
   float greedy = query.greedy_distance ( manip1->mat(), tree1, manip2->mat(), tree2, 0.1f );
   timer.stop();
   float dt5 = (float)timer.dt();

   sr_out << "collide:    " << (c/2) << " (" << dt1 << "s) "
                            <<  b << " (" << dt2 << "s) " << srnl;
   sr_out << "distance:   " << dist << " (" << dt3 << "s) "
                            << "dist(p1,p2):" << ::dist(dp1,dp2) << srnl;
   sr_out << "tolerance:  " << tole << " (" << dt4 << "s) "
                            << "dist(p1,p2):" << ::dist(tp1,tp2) << srnl;
   sr_out << "greedydist: " << greedy << " (" << dt5 << "s) " << srnl;
   sr_out<<srnl;

   // set line between closest points:
   SrLines& l = lines->shape();
   l.init();
   l.push_color ( SrColor::red );
   l.push_line ( dp1, dp2 );
   l.push_color ( SrColor::blue );
   l.push_line ( tp1, tp2 );

   // set colors of colliding triangles:
   model1->shape().Fm.setall(0);
   model2->shape().Fm.setall(0);
   for ( i=0; i<c; i+=2 )
    { model1->shape().Fm[query.colliding_pairs()[i]]=1;
      model2->shape().Fm[query.colliding_pairs()[i+1]]=1;
    }
 }

void manip_cb ( SrSnManipulator* manip, const SrEvent& e, void* udata )
 {
   Data* d = (Data*)udata;
   d->measure();
 }

int bv_main ( int argc, char **argv )
 {
   SrViewer* viewer = new SrViewer ( 385, 50, 500, 350 );

   viewer->label ( "PQP Collision, Distance and Tolerance queries in SR" );

   SrSnManipulator* manip1 = load ( "../data/knot.srm" );
   SrSnManipulator* manip2 = load ( "../data/sphere.srm" );
   SrSnLines* lines = new SrSnLines;

   manip1->translation ( SrVec(-10,0,0) );
   manip2->translation ( SrVec( 15,0,0) );

   Data* d = new Data ( manip1, manip2, lines );
   d->measure ();

   manip1->callback ( manip_cb, d );
   manip2->callback ( manip_cb, d );

   SrSnGroup* g = new SrSnGroup;
   g->add ( manip1 );
   g->add ( manip2 );
   g->add ( lines );

   viewer->root ( g );
   //viewer->menu_cmd ( SrViewer::CmdAxis );
   viewer->view_all();
   viewer->show();

   Fl::run();

   return 0;
 }
