  
# include <stdlib.h>

# include <FL/Fl.H>

# include <SR/sr_var.h>
# include <SR/sr_exp_table.h>

class VarTable : public SrExpTableVarManager
 { public :
    SrArray<SrVar*> varset;
    static VarTable *pt; // pointer to the first and unique instance of VarTable
   public :
    VarTable ()
     { pt = this;
       pt->ref ();
     }

    virtual ~VarTable ()
     { while ( varset.size()>0 ) delete varset.pop();
     };

    virtual int get_variable_index ( const char* var_name )
     { for ( int i=0; i<varset.size(); i++ )
        if ( sr_compare(varset[i]->name(),var_name)==0 ) return i;
       return -1;
     }

    virtual double get_variable_value ( int var_id, int array_id )
     { if ( var_id<0 || var_id>=varset.size() ) return 0.0;
       SrVar& v = *varset[var_id];
       array_id = SR_BOUND ( array_id, 0, v.size() );
       if ( v.type()=='i' )
        return (double) v.geti(array_id);
       else if ( v.type()=='f' )
        return (double) v.getf(array_id);
       else return 0;
     }

    void push ( const char *n=0, double val=0.0 )
     { varset.push() = new SrVar(n,'f');
       varset.top()->set ( val );
     }

    void update ( const SrVar& v )
     { int i = get_variable_index ( v.name() );
       if ( i==-1 )
        push ( v );
       else
        varset[i]->init ( v );
     }

    void push ( const SrVar& v )
     { varset.push() = new SrVar(v);
     }

    void list ()
     { sr_out << srnl;
       for ( int i=0; i<varset.size(); i++ )
       sr_out << i << ": " << *varset[i] << srnl;
     }
 };

VarTable* VarTable::pt = 0;

void set_var ( const char* input )
 {
   SrString s(input+2); // skip the ".v"
   s.append ( ";" );
   SrInput inp ( s );
   SrVar v;
   inp >> v;
   sr_out<<s<<" : "<<v<<srnl;
   VarTable::pt->update(v);
 }

void read_exp ( SrInput &in )
 {
   SrExpTable t;
   SrExpTable::Msg msg;
   int line;
   double result;

   t.set_var_manager ( VarTable::pt );

   msg = t.parse ( in, line );
   if ( msg!=SrExpTable::MsgOk ) 
    { sr_out << "\nParser error " << (int)msg << " in line " << line << ": \n"
             << SrExpTable::msg_desc(msg) << srnl;
      return;
    }

   msg = t.evaluate ( result );
   if ( msg!=SrExpTable::MsgOk ) 
    { sr_out << "\nEvaluation error : [" << SrExpTable::msg_desc(msg) << "].\n";
      return;
    }

   SrVar v ( "lr", 'f' );
   v.set ( result );
   VarTable::pt->update(v);

   sr_out << result << srnl;
 }

void help ()
 {
   sr_out << srnl;
   sr_out << "The prompt is ready to evaluate an expression.\n" ;
   sr_out << "Otherwise, the following commands are available:\n" ;
   sr_out << ".h - Display this help message\n";
   sr_out << ".q - Quit this prompt program\n";
   sr_out << ".f - List pre-defined evaluator functions\n";
   sr_out << ".s - List evaluator syntax\n";
   sr_out << ".l - List entered variables\n";
   sr_out << ".v - Enter a new variable\n";
 }

void list_pre_def ()
 {
   int i=0, col=4;
   sr_out << srnl;
   while ( true )
    { const char *func = SrExpTable::function_name(i++);
      if ( !func ) break;
      sr_out << func << (char)(i%col==0? '\n':'\t');
    }

   if ( (i-1)%col ) sr_out << srnl;
 }

void list_syntax ()
 {
   sr_out << srnl;
   sr_out << "Precedence: ^% /*: +- ><~= &|\n" <<
             "x^y = x raised to y\n" <<
             "x%y = x percent of y\n" <<
             "x:y = x mod y.\n" <<
             "&|~= mean: and, or, difference, equal\n" <<
             "! is a logical unary operator\n" <<
             "< > <= >= true false are supported\n";
 }

void enter_cb ( const char* inputs, void* /*udata*/ )
 {
   char cmd = inputs[0];

   if ( cmd=='.' )
    { cmd = inputs[1];

      if ( cmd=='h' )  help ();
       else
      if ( cmd=='f' ) list_pre_def ();
       else
      if ( cmd=='s' ) list_syntax ();
       else
      if ( cmd=='q' ) exit(0);
       else
      if ( cmd=='l' ) VarTable::pt->list ();
       else
      if ( cmd=='v' ) set_var (inputs);
       else
      sr_out << "\nUndefined cmd!\n";
    }
   else
    { SrInput in(inputs);
      read_exp(in);
    }
 }

# include <FL/Fl_Window.H>
class Fl_Input;
class Fl_Browser;

class FluShell : public Fl_Window
 { private :
    Fl_Input *_input;
    Fl_Browser *_output;
    void (*_enter_cb) (const char*,void*);
    void* _udata;
    int _length;
   private :
    friend void input_enter_cb ( Fl_Input* i, void* win );
   public :
    FluShell ( int x, int y, int w, int h, const char *title=0 );
    FluShell ( int w, int h, const char *title=0 );
    void callback ( void (*cb)(const char*,void*), void* udata=0 )
         { _enter_cb=cb; _udata=udata; }
    /*! default is 50 lines. */
    void history_length ( int l ) { _length=l; }
    void user_data ( void* udata ) { _udata=udata; }
    void hide_input ();
    void output ( const char *s );
   public :
    /*! This virtual method is invoked whenever something is entered in the shell.
        Its implementation simply calls the callback function registered with callback(). */
    virtual void text_entered ( const char* s, void* udata );
 };

void shell_output ( const char* s, void* udata )
 {
   FluShell *shell = (FluShell*)udata;
   shell->output ( s );
 }

int exp_main ( int, char ** )
 {
   new VarTable;
   VarTable::pt->push( "e", sre );
   VarTable::pt->push( "pi", srpi );

   FluShell *shell = new FluShell ( 500, 300, "SR Evaluator Shell" );
   shell->callback ( enter_cb, shell );
   shell->show();

   sr_out.init ( shell_output, (void*)shell );

   sr_out << "\n";
   sr_out << "Sr Expression Prompt, M. Kallmann\n";
   sr_out << "Enter '.h' for help.\n";

   Fl::run ();

   return 0;
 }

//=========================== FluShell =====================================

# include <stdlib.h>

# include <FL/Fl.H>
# include <FL/Fl_Browser.H>
# include <FL/Fl_Input.H>

void input_enter_cb ( Fl_Input* i, void* win )
 {
   FluShell* shell = ((FluShell*)win);
   shell->text_entered ( i->value(), shell->_udata );
   shell->_input->value("");
 }

void set ( int w, int h, Fl_Input*& input, Fl_Browser*& output, FluShell *win )
 {
   if ( h<50 ) h=50;
   int dy = 25;
   int inpy = h-dy;
   input = new Fl_Input ( 0, inpy, w, dy );
   input->callback ( (Fl_Callback*)input_enter_cb, win );
   input->when(FL_WHEN_ENTER_KEY_ALWAYS);
   input->textfont(4);
   output = new Fl_Browser ( 0, 0, w, inpy );
   output->textfont(4);
 }

FluShell::FluShell ( int x, int y, int w, int h, const char* title )
         :Fl_Window ( x, y, w, h, title )
 {
   _udata = 0;
   _enter_cb = 0;
   _length = 50;
   set ( w, h, _input, _output, this );
   Fl_Group::current()->resizable(_output);
   end();
 }

FluShell::FluShell ( int w, int h, const char* title )
         :Fl_Window ( w, h, title )
 {
   _udata = 0; 
   _enter_cb = 0;
   _length = 50;
   set ( w, h, _input, _output, this );
   Fl_Group::current()->resizable(_output); // the 'Fl_Group::' is important!
   end();
 }

void FluShell::hide_input ()
 {
   _input->hide();
   _output->resize ( 0, 0, _output->w(), _output->h()+25 );
 }

# define GAP 32
static char* buffer=0; // line buffer
static int bufcur=0;
static int bufcap=0;

void buffer_push ( char c )
 {
   if ( bufcur>=bufcap )
    { bufcap += GAP;
      buffer = (char*) realloc ( buffer, bufcap );
    }
   buffer[bufcur++]=c;
 }

void FluShell::output ( const char *s )
 {
   while ( *s )
    { buffer_push(*s);

      if ( *s=='\n' )
       { buffer_push(0);

         _output->add ( buffer );
         bufcur=0;

         int extra = _output->size()-_length;
         if ( extra>0 ) _output->remove ( 1 );
       }
      s++;
    }

   _output->value ( _output->size() );
   _output->deselect();
 }


void FluShell::text_entered ( const char* s, void* udata )
 {
   if ( _enter_cb ) _enter_cb ( s, udata );
 }

