
# include <stdlib.h>

# include <FL/fl_ask.H>
# include <FL/fl_file_chooser.H>

# include <SR/sr_polygon.h>
# include <SR/sr_fl.h>

# include "semesh_main_win.h"
# include "semesh_mesh.h"
# include "semesh_main.h"

//# define SR_USE_TRACE1  // event tracing
# include <SR/sr_trace.h>

MainWin::MainWin ()
 {
 }

void MainWin::show ()
 {
   window->show();
 }

void MainWin::render ()
 {
   viewer->visibility ( draw_vtx->value()? true:false,
                        draw_edg->value()? true:false,
                        draw_fac->value()? true:false,
                        draw_red_parm->value()? true:false,
                        draw_blue_parm->value()? true:false,
                        draw_point_parm->value()? true:false );
   viewer->render ();
 }

//=================================== events ===============================

// callback called from color window:
void apply_elem_color ( char elem )
 {
   Fl_Button *b = elem=='v' ? TheColorWin->vtx :
                  elem=='e' ? TheColorWin->edg : TheColorWin->fac;

   SrColor c;
   Fl::get_color ( b->color(), c.r, c.g, c.b );

   if ( elem=='v' )
    { TheRedSymEdge->vtx()->c = c;
    }
   else if ( elem=='e' )
    { TheRedSymEdge->edg()->c = c;
    }
   else if ( elem=='f' )
    { TheRedSymEdge->fac()->c = c; 
    }

   TheMainWin->viewer->draw_mesh ();
   TheMainWin->render ();
 }

void MainWin::event ( FileEvent e )
 {
   switch ( e )
    { case evFileOpen :
       { char *file = fl_file_chooser ( "Open File", "*.se", 0 );
         if ( !file ) return;
         SrInput inp ( fopen(file,"rt") );
         if ( !inp.valid() ) { fl_alert ( "Could not open %s!", file ); break; }
         if ( !TheMesh->load(inp) ) { fl_alert ( "File %s is not valid!", file ); break; }
         TheRedSymEdge = TheMesh->first();
         TheBlueSymEdge = TheRedSymEdge->sym();
         viewer->draw_mesh ();
         viewer->draw_parms ();
         render ();
       } break;

      case evFileSave :
       { char *file = fl_file_chooser ( "Save File", "*.se", 0 );
         if ( !file ) return;
         FILE* f = fopen(file,"wt");
         if ( !f ) { fl_alert ( "Could not save %s!", file ); break; }
         SrOutput o ( f );
         if ( !TheMesh->save(o) ) fl_alert ( "Error while saving %s!", file );
       } break;

      case evFileExit : exit(0); break;
    }
 }

void MainWin::event ( MeshEvent ev )
 { 
   Mesh* mesh = TheMesh;

   switch ( ev )
    { case evMeshInfo : 
       fl_message ( "Vertices:%d Edges:%d Faces:%d Genus:%d\nMean Vertex Degree:%f",
                    mesh->vertices(), mesh->edges(), mesh->faces(),
                    mesh->genus(), mesh->mean_vertex_degree() );
       break;

      case evMeshColors : TheColorWin->window->show(); break;

      case evMeshInit : 
       TheRedSymEdge = mesh->init ( SrPnt(0.5f,0,0), SrPnt(-0.5f,0,0) );
       TheBlueSymEdge = TheRedSymEdge->sym();
       viewer->draw_mesh ();
       viewer->draw_parms ();
       render ();
       break;

      case evMeshCheck : 
       { Mesh::ChkMsg m = TheMesh->check_all();
         if ( m==Mesh::ChkNxtError )
          fl_alert ( "BUG: Inconsistency in nxt() operator found!" );
         else if ( m==Mesh::ChkRotError )
          fl_alert ( "BUG: Inconsistency in rot() operator found!" );
         else if ( m==Mesh::ChkVtxError )
          fl_alert ( "BUG: Inconsistency in vertex referencing!" );
         else if ( m==Mesh::ChkEdgError )
          fl_alert ( "BUG: Inconsistency in edge referencing!" );
         else if ( m==Mesh::ChkFacError )
          fl_alert ( "BUG: Inconsistency in face referencing!" );
         else
          fl_message ( "Check OK!" );
       } break;

      case evMeshInvertFaces : 
       TheMesh->invert_faces (); 
       viewer->draw_mesh ();
       viewer->draw_parms ();
       render ();
       break;

      case evMeshFindEdge : 
      case evMeshFindEdgeBlue : 
       { Edge *e, *ei, *fe=0;
         SrPnt mp, p;
         float mind=0, d=0;
         p = *ThePoint;
         if ( TheMesh->empty() ) break;
         e = ei = TheMesh->first()->edg();
         do { mp = ( e->se()->vtx()->p + e->se()->nxt()->vtx()->p ) / 2.0f;
              d = dist ( mp, p );
              if ( fe==0 || d<mind ) { fe=e; mind=d; }
              e = e->nxt();
            } while ( e != ei );
         if ( ev==evMeshFindEdge ) TheRedSymEdge = fe->se();
           else TheBlueSymEdge = fe->se();
         viewer->draw_parms ();
         render ();
       } break;
    }
 }

void MainWin::event ( DrawEvent e )
 {
   render ();
 }

void MainWin::event ( OpEvent e )
 {
   SrPnt &p = *ThePoint;
   SymEdge *s1 = TheRedSymEdge;
   SymEdge *s2 = TheBlueSymEdge;
   Mesh* mesh = TheMesh;
  
   // if deleting a vertex, save coordinates in p:
   switch ( e )
    { case evKev  :
      case evKev2 : 
      case evEcol :
      case evDelv : p=s1->vtx()->p; break;
    }

   SR_TRACE1 ( "Operator Event..." );
   switch ( e )
    { case evMev  : s1=mesh->mev(s1,p); break;
      case evMev2 : s1=mesh->mev(s1,s2,p); break;
      case evMef  : s1=mesh->mef(s1,s2); break; // check ccw!
      case evKev  : s1=mesh->kev(s1); break;
      case evKev2 : s1=mesh->kev(s1,&s2); break; 
      case evKef  : s1=mesh->kef(s1,&s2); break;
      case evEcol : s1=mesh->ecol(s1,&s2); break;
      case evVsplit : s1=mesh->vsplit(s1,s2,p); break;
      case evFlip : s1=mesh->flip(s1); break;
      case evAddv : s1=mesh->addv(s1,p); break;
      case evDelv : s1=mesh->delv(s1); break;
      case evUpdVtx : s1->vtx()->p = p; break;
      case evUpdVtxBlue : s2->vtx()->p = p; break;
    }
   SR_TRACE1 ( "Operator Executed." );

   if ( !s1 ) // error
    { fl_alert ( "Operator could not be performed:\n%s",
                 Mesh::translate_op_msg(mesh->last_op_msg()) );
    }
   else
    { TheRedSymEdge=s1;
      TheBlueSymEdge=s2;
      viewer->draw_mesh ();
      viewer->draw_parms ();
      render ();
    }
 }

void MainWin::event ( TravEvent e )
 {
   SymEdge*& s1 = TheRedSymEdge;
   SymEdge*& s2 = TheBlueSymEdge;

   SR_TRACE1 ( "Traverse Event..." );

   switch ( e )
    { case evRedNxt : s1=s1->nxt(); break;
      case evRedPri : s1=s1->pri(); break;
      case evRedSym : s1=s1->sym(); break;
      case evRedRot : s1=s1->rot(); break;
      case evRedRet : s1=s1->ret(); break;
      case evBlueNxt : s2=s2->nxt(); break;
      case evBluePri : s2=s2->pri(); break;
      case evBlueSym : s2=s2->sym(); break;
      case evBlueRot : s2=s2->rot(); break;
      case evBlueRet : s2=s2->ret(); break;
    }
   viewer->draw_parms();
   render ();
 }

void MainWin::event ( HelpEvent e )
 {
   switch ( e )
    { case evHelpAbout :
       { static char* text =
          "This application demonstrates the concepts of the Symmetrical Edge data\n"
          "structure implemented in the SeMesh class of the SE toolkit.\n\n"
          "The structure joins Guibas & Stolfi ideas with Mntyl Euler operators.\n"
          "It is based on circular lists of vertices, edges and faces, plus the\n"
          "nxt and rot connectivity information.\n\n"
          "References:\n"
          "  1. M. Kallmann, and D. Thalmann, Star-Vertices: A Compact Representation\n"
          "     for Planar Meshes with Adjacency Information, Journal of Graphics Tools,\n"
          "     2001. (In the comparisons table, the symedge was used and called as the\n"
          "     simplified quad-edge structure )\n"
          "  2. M. Mntyl, An Introduction to Solid Modeling, Computer Science Press,\n"
          "     Maryland, 1988.\n"
          "  3. L. Guibas and J. Stolfi, Primitives for the Manipulation of General\n"
          "     Subdivisions and the Computation of Voronoi Diagrams, ACM Transaction\n"
          "     on Graphics, 4, 75-123, 1985.\n";

          fl_text_output ( "SeMesh About", text );
       } break;
    }
 }
