
# include <SR/sr.h>
//# include <SR/var_compact.h>

/* VarCompact class was the old built-in SrVar class, which was replaced
   by a new one using SrArray.
   This old version is more memory compact, but
   much less efficient for changing dynamically array sizes.
   Also, the idea of VarCompact is more like a programming language var,
   while the new SrVar is more for general data storage and description */

//========================= var_compact.h ===============================

# ifndef VAR_COMPACT_H
# define VAR_COMPACT_H

/** \file var_compact.h 
 * a multi type variable
 */

# include <SR/sr.h>
# include <SR/sr_output.h>

/*! \class VarCompact var_compact.h
    \brief generic type variable

    VarCompact keeps generic type and multi dimension variable values. 
    The type and dimension size are given with method init().
    Methods set() and get() can be used to access/change the variable values,
    and they check if the correct types are being employed.
    To change the type of a VarCompact, init() or clone() must be used.
    Type casts are performed automatically with method assign(). */
class VarCompact
 { private :
    union Data 
     { bool b;   char* B; 
       int i;    int* I;
       float f;  float* F;
       double d; double* D;
       char* s;  srcharpt* S;
     };

    char   _type;
    Data   _data; // SrArray<Data> _arraydata;
    sruint _size;
    char*  _name;
    void _free_data ();
    void _free_name ();
    void _convert ( char t1, char t2, const VarCompact::Data& d1, VarCompact::Data& d2 );
   public :    

    /*! Constructs VarCompact with type 'i', value 0, and name "" */
    VarCompact ();

    /*! Copy constructor */
    VarCompact ( const VarCompact& v );

    /*! Copy constructor from one element of a multidimensional VarCompact v */
    VarCompact ( const VarCompact& v, sruint index );

    /*! Deletes all used internal data */
    ~VarCompact ();

    /*! Sets a name to be kept by VarCompact. Null can be passed to clear the name. */
    void name ( const char* n );

    /*! Returns the associated name. "" is returned if no name was set. */
    const char* name () const;

    /*! Inits VarCompact with the given type. Types are specified by the first
        letter of: bool, int, float, double, string. The variable value can then
        be set with one of the set methods */
    void init ( char type );

    /*! Inits VarCompact with the given type and size. Types are specified by the
        first letter in upper case of: bool, int, float, double, string.
        The size must be >0 */
    void init ( char type, sruint size );

    /*! Makes VarCompact be identical to v, same as operator= */
    void init ( const VarCompact& v );

    /*! Get the value at the given position index, and put it in v. */
    void init ( const VarCompact& v, sruint index );

    /*! Returns the actual array size of the variable */
    sruint size () const { return _size; }

    /*! Returns one letter corresponding to the type of the variable: bifdsBIFDS */
    char type () const { return _type; }

    void set ( bool b );           //!< set a bool value
    void set ( int i );            //!< set an int value
    void set ( float f );          //!< set a float value
    void set ( double d );         //!< set a double value
    void set ( const char* s );    //!< set a string value

    void set ( sruint index, bool b );        //!< set a bool value at a given position
    void set ( sruint index, int i );         //!< set an int value at a given position
    void set ( sruint index, float f );       //!< set a float value at a given position
    void set ( sruint index, double d );      //!< set a double value at a given position
    void set ( sruint index, const char* s ); //!< set a string value at a given position

    bool getb () const;                       //!< get the bool value
    int geti () const;                        //!< get the int value
    float getf () const;                      //!< get the float value
    double getd () const;                     //!< get the double value
    const char* gets () const;                //!< get the string value

    bool getb ( sruint index ) const;         //!< get the bool value at the given position
    int geti ( sruint index ) const;          //!< get the int value at the given position
    float getf ( sruint index ) const;        //!< get the float value at the given position
    double getd ( sruint index ) const;       //!< get the double value at the given position
    const char* gets ( sruint index ) const;  //!< get the string value at the given position

    /*! Assigns the value of v to VarCompact performing type casts if needed */
    void assign ( const VarCompact& v );

    /*! Makes VarCompact be identical to v, same as init(v) */
    VarCompact& operator= ( const VarCompact& v );

    /*! Outputs VarCompact */
    friend SrOutput& operator<< ( SrOutput& o, const VarCompact& v );
 };

//================================ End of File =================================================

# endif  // VAR_COMPACT_H

//================================ var_compact.cpp =================================================


# include <string.h>
# include <stdlib.h>

//# include <SR/var_compact.h>
# include <SR/sr_output.h>

//======================================== VarCompact =======================================

//--------------------------- private methods ---------------------------

void VarCompact::_free_data ()
 {
   switch (_type)
    { case 's' : delete[] _data.s; break;
      case 'B' : delete[] _data.B; break;
      case 'I' : delete[] _data.I; break;
      case 'F' : delete[] _data.F; break;
      case 'D' : delete[] _data.D; break;
      case 'S' : { for ( sruint i=0; i<_size; i++ ) delete[]_data.S[i]; }
                 delete[] _data.S; break;
    }
 }
 
void VarCompact::_free_name ()
 {
   delete _name;
   _name = 0;
 }

void VarCompact::_convert ( char t1, char t2, const VarCompact::Data& d1, VarCompact::Data& d2 )
 {
   t1 = SR_LOWER(t1);
   t2 = SR_LOWER(t2);

   if ( t2=='s' )
    { char buf[60];
      switch ( t1 )
       { case 'b' : strcpy ( buf, d1.b?"true":"false" ); break;
         case 'i' : sprintf ( buf, "%d", d1.i ); break;
         case 'f' : sprintf ( buf, "%f", d1.f ); break;
         case 'd' : sprintf ( buf, "%f", d1.d ); break;
         case 's' : strcpy ( buf, d1.s ); break;
       }
      sr_string_set ( d2.s, buf );
      return;
    };
//printf(" %c%c \n",t1,t2);
   switch ( t1 )
    { case 'b' : switch ( t2 )
       { case 'b' : d2.b=d1.b; break;
         case 'i' : d2.i=(int)d1.b; break;
         case 'f' : d2.f=(float)d1.b; break;
         case 'd' : d2.d=(double)d1.b; break;
       } break;

      case 'i' : switch ( t2 )
       { case 'b' : d2.b=d1.i?1:0; break;
         case 'i' : d2.i=d1.i; break;
         case 'f' : d2.f=(float)d1.i; break;
         case 'd' : d2.d=(double)d1.i; break;
       } break;

      case 'f' : switch ( t2 )
       { case 'b' : d2.b=d1.f==0?0:1; break;
         case 'i' : d2.i=(int)d1.f; break;
         case 'f' : d2.f=d1.f; break;
         case 'd' : d2.d=(float)d1.f; break;
       } break;

      case 'd' : switch ( t2 )
       { case 'b' : d2.b=d1.d==0?0:1; break;
         case 'i' : d2.i=(int)d1.d; break;
         case 'f' : d2.f=(float)d1.d; break;
         case 'd' : d2.d=d1.d; break;
       } break;

      case 's' : switch ( t2 )
       { case 'b' : d2.b=sr_compare(d1.s,"true")==0?1:0; break;
         case 'i' : d2.i=atoi(d1.s); break;
         case 'f' : d2.f=(float)atof(d1.s); break;
         case 'd' : d2.d=(double)atof(d1.s); break;
       } break;
    }
 }

//--------------------------- public methods ---------------------------

VarCompact::VarCompact ()
 {
   _type = 'i';
   _data.i = 0;
   _size = 0;
   _name = 0;
 }

VarCompact::VarCompact ( const VarCompact& v )
 {
   _type = 'i';
   _data.i = 0;
   _size = 0;
   _name = 0;
   init ( v );
 }

VarCompact::VarCompact ( const VarCompact& v, sruint index )
 {
   _type = 'i';
   _data.i = 0;
   _size = 0;
   _name = 0;
   init ( v, index );
 }

VarCompact::~VarCompact ()
 {
   _free_name ();
   _free_data ();
 }

void VarCompact::name ( const char* n )
 {
   sr_string_set ( _name, n );
 }

const char* VarCompact::name () const
 {
   return _name? _name:"";
 }

void VarCompact::init ( char type )
 {
   _free_data ();
   _type = type;
   _size = 0;
   switch (type)
    { case 'b' : _data.b=false; break;
      case 'i' : _data.i=0; break;
      case 'f' : _data.f=0; break;
      case 'd' : _data.d=0; break;
      case 's' : _data.s=0; break;
      default : sr_out.fatal_error ( "Invalid call to VarCompact::init(type)" ); break;
    }
 }

void VarCompact::init ( char type, sruint size )
 {
   sruint i;
   _free_data ();
   _type = type;
   _size = size;
   SR_ASSERT ( _size>0 );
   switch ( type )
    { case 'B' : _data.B = new char[size];     for ( i=0; i<size; i++ ) _data.B[i]=0; break;
      case 'I' : _data.I = new int[size];      for ( i=0; i<size; i++ ) _data.I[i]=0; break;
      case 'F' : _data.F = new float[size];    for ( i=0; i<size; i++ ) _data.F[i]=0; break;
      case 'D' : _data.D = new double[size];   for ( i=0; i<size; i++ ) _data.D[i]=0; break;
      case 'S' : _data.S = new srcharpt[size]; for ( i=0; i<size; i++ ) _data.S[i]=0; break;
      default : sr_out.fatal_error ( "Invalid call to VarCompact::init(type,size)" ); break;
    }
 }

void VarCompact::init ( const VarCompact& v )
 {
   _free_data ();
   name ( v.name() );
   sruint s = v.size();

   if ( s==0 )
    init ( v.type() );
   else
    init ( v.type(), s );

   assign ( v );
 }

void VarCompact::init ( const VarCompact& v, sruint index )
 {
   SR_ASSERT ( index<v._size );
   name ( v.name() );
   char t = SR_LOWER(v._type);
   init ( t );
   switch ( t )
    { case 'b' : _data.b=v._data.B[index]? true:false; break;
      case 'i' : _data.i=v._data.I[index]; break;
      case 'f' : _data.f=v._data.F[index]; break;
      case 'd' : _data.d=v._data.D[index]; break;
      case 's' : sr_string_set(_data.s,v._data.S[index]); break;
    }
 }

//--------------------------- set single value ---------------------------

void VarCompact::set ( bool b )
 {
   SR_ASSERT ( _type=='b' );
   _data.b = b;
 }

void VarCompact::set ( int i )
 {
   SR_ASSERT ( _type=='i' );
   _data.i = i;
 }

void VarCompact::set ( float f )
 {
   SR_ASSERT ( _type=='f' );
   _data.f = f;
 }

void VarCompact::set ( double d )
 {
   SR_ASSERT ( _type=='d' );
   _data.d = d;
 }

void VarCompact::set ( const char* s )
 {
   SR_ASSERT ( _type=='s' );
   sr_string_set ( _data.s, s );
 }

//--------------------------- set indexed value ---------------------------

void VarCompact::set ( sruint index, bool b )
 {
   SR_ASSERT ( _type=='B' );
   SR_ASSERT ( index<_size );
   _data.B[index] = (char)b;
 }

void VarCompact::set ( sruint index, int i )
 {
   SR_ASSERT ( _type=='I' );
   SR_ASSERT ( index<_size );
   _data.I[index] = i;
 }

void VarCompact::set ( sruint index, float f )
 {
   SR_ASSERT ( _type=='F' );
   SR_ASSERT ( index<_size );
   _data.F[index] = f;
 }

void VarCompact::set ( sruint index, double d )
 {
   SR_ASSERT ( _type=='D' );
   SR_ASSERT ( index<_size );
   _data.D[index] = d;
 }

void VarCompact::set ( sruint index, const char* s )
 {
   SR_ASSERT ( _type=='S' );
   SR_ASSERT ( index<_size );
   sr_string_set ( _data.S[index], s );
 }

//--------------------------- get single value ---------------------------

bool VarCompact::getb () const
 {
   SR_ASSERT ( _type=='b' );
   return _data.b;
 }

int VarCompact::geti () const
 {
   SR_ASSERT ( _type=='i' );
   return _data.i;
 }

float VarCompact::getf () const
 {
   SR_ASSERT ( _type=='f' );
   return _data.f;
 }

double VarCompact::getd () const
 {
   SR_ASSERT ( _type=='d' );
   return _data.d;
 }

const char* VarCompact::gets () const
 {
   SR_ASSERT ( _type=='s' );
   return _data.s? _data.s:"";
 }

//--------------------------- get indexed value ---------------------------

bool VarCompact::getb ( sruint index ) const
 {
   SR_ASSERT ( _type=='B' );
   SR_ASSERT ( index<_size );
   return _data.B[index]? true:false;
 }

int VarCompact::geti ( sruint index ) const
 {
   SR_ASSERT ( _type=='I' );
   SR_ASSERT ( index<_size );
   return _data.I[index];
 }

float VarCompact::getf ( sruint index ) const
 {
   SR_ASSERT ( _type=='F' );
   SR_ASSERT ( index<_size );
   return _data.F[index];
 }

double VarCompact::getd ( sruint index ) const
 {
   SR_ASSERT ( _type=='D' );
   SR_ASSERT ( index<_size );
   return _data.D[index];
 }

const char* VarCompact::gets ( sruint index ) const
 {
   SR_ASSERT ( _type=='S' );
   SR_ASSERT ( index<_size );
   const char* st = _data.S[index];
   return st? st:"";
 }

void VarCompact::assign ( const VarCompact& v )
 {
   sruint i;
   sruint s = SR_MIN(v._size,_size);
   char t1 = v._type;
   char t2 = _type;  // data t1 => data t2

   if ( t1==t2 ) // no conversion, just copy
    { switch ( t2 )
       { case 'b' : _data.b=v._data.b; break;
         case 'i' : _data.i=v._data.i; break;
         case 'f' : _data.f=v._data.f; break;
         case 'd' : _data.d=v._data.d; break;
         case 's' : sr_string_set(_data.s,v._data.s); break;
         case 'B' : for ( i=0; i<s; i++ ) _data.B[i]=v._data.B[i]; break;
         case 'I' : for ( i=0; i<s; i++ ) _data.I[i]=v._data.I[i]; break;
         case 'F' : for ( i=0; i<s; i++ ) _data.F[i]=v._data.F[i]; break;
         case 'D' : for ( i=0; i<s; i++ ) _data.D[i]=v._data.D[i]; break;
         case 'S' : for ( i=0; i<s; i++ ) sr_string_set(_data.S[i],v._data.S[i]); break;
       }
    }
   else 
    { if ( _size==0 && v._size==0 ) // v value in VarCompact
       { _convert ( t1, t2, v._data, _data );
       }
      else if ( _size>0 && v._size>0 )
       { Data d1, d2;
         for ( i=0; i<s; i++ )
          { switch ( t1 )
             { case 'B' : d1.b=v._data.B[i]? true:false; break;
               case 'I' : d1.i=v._data.I[i]; break;
               case 'F' : d1.f=v._data.F[i]; break;
               case 'D' : d1.d=v._data.D[i]; break;
               case 'S' : d1.s=v._data.S[i]; break;
             }
            switch ( t2 )
             { case 'B' : d2.b=_data.B[i]? true:false; break;
               case 'I' : d2.i=_data.I[i]; break;
               case 'F' : d2.f=_data.F[i]; break;
               case 'D' : d2.d=_data.D[i]; break;
               case 'S' : d2.s=_data.S[i]; break;
             }
            _convert ( t1, t2, d1, d2 );
            switch ( t2 )
             { case 'B' : _data.B[i]=d2.b; break;
               case 'I' : _data.I[i]=d2.i; break;
               case 'F' : _data.F[i]=d2.f; break;
               case 'D' : _data.D[i]=d2.d; break;
               case 'S' : _data.S[i]=d2.s; break;
             }
          }
       }
      else if ( _size==0 && v._size>0 ) // v value in VarCompact
       { VarCompact x (v,0); 
         _convert ( SR_LOWER(t1), t2, x._data, _data );
       }
      else
       { char lt2 = SR_LOWER(t2);
         VarCompact x; x.init(lt2); 
         _convert ( t1, lt2, v._data, x._data );
         switch ( t2 )
          { case 'B' : _data.B[0]=x._data.b; break;
            case 'I' : _data.I[0]=x._data.i; break;
            case 'F' : _data.F[0]=x._data.f; break;
            case 'D' : _data.D[0]=x._data.d; break;
            case 'S' : _data.S[0]=x._data.s; x._data.s=0; break;
          }
       }
    }
 }

VarCompact& VarCompact::operator= ( const VarCompact& v )
 {
   init ( v );
   return *this;
 }

//============================== friends ========================================

SrOutput& operator<< ( SrOutput& o, const VarCompact& v )
 {
   sruint i;
   sruint s = v._size;
   char lt = SR_LOWER(v._type);
   switch ( lt )
    { case 'b' : o<<"bool"; break;
      case 'i' : o<<"int"; break;
      case 'f' : o<<"float"; break;
      case 'd' : o<<"double"; break;
      case 's' : o<<"string"; break;
    }

   if ( v._name ) o << ' ' << v.name();
   if ( s>1 ) o << '[' << s << "]";
   o << " = ";

   s--;
   switch ( v._type )
    { case 'b' : o<<v._data.b; break;
      case 'i' : o<<v._data.i; break;
      case 'f' : o<<v._data.f; break;
      case 'd' : o<<v._data.d; break;
      case 's' : o<<'"'<<v.gets()<<'"'; break;
      case 'B' : for ( i=0; i<=s; i++ ) { o<<(v._data.B[i]?true:false); if (i<s) o<<srspc; } break;
      case 'I' : for ( i=0; i<=s; i++ ) { o<<v._data.I[i]; if (i<s) o<<srspc; } break;
      case 'F' : for ( i=0; i<=s; i++ ) { o<<v._data.F[i]; if (i<s) o<<srspc; } break;
      case 'D' : for ( i=0; i<=s; i++ ) { o<<v._data.D[i]; if (i<s) o<<srspc; } break;
      case 'S' : for ( i=0; i<=s; i++ ) { o<<'"'<<v.gets(i)<<'"'; if (i<s) o<<srspc; } break;
    }

   return o<<';';
 }

/*
SrInput& operator>> ( SrInput& in, VarCompact& box )
 {
   return in >> box.a >> box.b;
 }
*/

//================================ End of File =================================================

int var_compact_main ( int, char ** )
 {
   VarCompact a, b, c, d, e;

   a.init ( 'S', 4 ); a.name("var_a");
   a.set(0,"one"); a.set(1,"two"); a.set(2,"true"); a.set(3,"true");

   b.init ( 'B', 7 ); b.name("var_b"); b.set(1,true);

   c.init ( 'D', 2 ); c.name("var_c"); c.set(1,3.1415);

   d.init ( 's' );  d.name("var_d"); 

   e.init ( 's' );  d.name("var_d"); 

   a.assign(c);
   c.assign(a);

   sr_out<<a<<srnl;
   sr_out<<b<<srnl;
   sr_out<<c<<srnl;
   sr_out<<d<<srnl;
   sr_out<<e<<srnl;

   return 0;
 }
